// machine.h 
//  Data structures for simulating the execution of user programs
//  running on top of Nachos.
//
//  User programs are loaded into "mainMemory"; to Nachos,
//  this looks just like an array of bytes.  (Of course, the Nachos
//  kernel is in memory too -- but in most machines these days,
//  the kernel is loaded into a separate memory region from user
//  programs, and accesses to kernel memory are not translated or paged.
//
//  In Nachos, user programs are executed one instruction at a time, 
//  by the simulator.  Each memory reference is translated, checked
//  for errors, etc.
//
//  DO NOT CHANGE -- part of the machine emulation
//
// Copyright (c) 1992-1993 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#ifndef MACHINE_H
#define MACHINE_H

#ifndef HOST_DOS
#include "copyright.h"
#include "utility.h"
#include "tlb.h"
#include "disk.h"
#else
#include "copyrght.hpp"
#include "utility.hpp"
#include "tlb.hpp"
#include "disk.hpp"
#endif

// Definitions related to the size, and format of user memory

#define PageSize    SectorSize  // set the page size equal to
                    // the disk sector size, for
                    // simplicity

#define NumPhysPages    32
#define MemorySize  (NumPhysPages * PageSize)

#ifdef VM           // if VM, make the TLB small for realism
#define TLBSize     4
#else               // for now, make the TLB big for simplicity
#define TLBSize     NumPhysPages 
#endif

// User program behavior can cause control to transfer to the Nachos kernel.
// This is the list of exceptions that a user program can cause. 
enum ExceptionType { SyscallException,      // A program executed a system call.
             NoTLBEntryException,   // No translation found in the TLB
             ReadOnlyException,     // Write attempted to page marked 
                        // "read-only"
             BusErrorException,     // Translation resulted in an 
                        // invalid physical address
             AddressErrorException, // Unaligned reference
             OverflowException,     // Integer overflow in add or sub.
             IllegalInstrException, // Unimplemented or reserved instr.
             
             NumExceptionTypes
};

// User program CPU state.  The full set of CPU registers, plus a few
// more because we need to be able to start/stop a user program between
// any two instructions (thus we need to keep track of things like load
// delay slots, etc.)
#define StackReg    29  // user's stack pointer
#define RetAddrReg  31  // holds return address for procedure calls
#define NumGPRegs   32  // 32 general purpose registers on MIPS
#define HiReg       32  // double register to hold multiply result
#define LoReg       33
#define PCReg       34  // current program counter
#define NextPCReg   35  // next program counter (eg, branch delay) 
#define PrevPCReg   36  // previous program counter
#define LoadReg     37  // The register target of a delayed load.
#define LoadValueReg    38  // The value to be loaded by a delayed load.
#define BadVAddrReg 39  // the virtual address on an exception

#define NumTotalRegs    40

// The following class defines the simulated host workstation hardware, as 
// seen by user programs -- the CPU registers, main memory, etc.
// User programs can't tell that they are running on our simulator
// or on the real hardware (except we don't support floating point 
// instructions, and the system call interface to Nachos is not the
// same as UNIX (10 system calls in Nachos vs. 200 in UNIX!)
//
// The procedures in this class are split between machine.cc and mipssim.cc.
class Machine {
  public:
    Machine(bool debug);    // initialize the simulation of the hardware
                // for running user programs
    ~Machine();         // de-allocate the data structures

    void Run();         // Run a user program

    void OneInstruction();  // Run one instruction of a user program.
    void DelayedLoad(int nextReg, int nextVal);     
                // Do a pending delayed load (modifying a reg)
    
    void RaiseException(ExceptionType which, int badVAddr);
                // Trap to the Nachos kernel, because of a
                // system call or other exception

    void Debugger();        // invoke the user program debugger
    void DumpState();       // print the user CPU and memory state 

    int ReadRegister(int num);  // read the contents of a CPU register
    void WriteRegister(int num, int value);
                // store a value into a CPU register

// Routines for converting Words and Short Words to and from the
// simulated machine's format of little endian.  If the host machine
// is little endian (as it is on the DEC MIPS), these end up being NOPs.

    unsigned int WordToHost(unsigned int word);
    unsigned short ShortToHost(unsigned short shortword);
    unsigned int WordToMachine(unsigned int word);
    unsigned short ShortToMachine(unsigned short shortword);
    
// These data structures are public, for convenience.
// Note that *all* communication between the user program and the kernel 
// is in terms of these data structures.
    char *mainMemory;       // physical memory to store user program,
                // code and data, while executing
    int registers[NumTotalRegs]; // CPU registers, for executing user programs
    TLB *tlb;           // translation lookaside buffer, for managing
                // the mapping of virtual addresses in the
                // user program, to physical addresses relative
                // to the start of "mainMemory"

  private:
    bool singleStep;        // drop back into the debugger after each
                // simulated instruction
    int runUntilTime;       // drop back into the debugger when simulated
                // time reaches this value
};

extern void ExceptionHandler(ExceptionType which);
                // Entry point into Nachos for handling
                // user system calls and exceptions
                // Defined in exception.cc

#endif // MACHINE_H
