// tlb.cc 
//	Routines to translate virtual addresses to physical addresses.
//	Software sets up a table of legal translations.  We look up
//	in the table on every memory reference to find the true physical
//	memory location.
//
// DO NOT CHANGE -- part of the machine emulation
//
// Copyright (c) 1992-1993 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include "copyright.h"
#include "tlb.h"
#include "addrspace.h"
#include "system.h"

//----------------------------------------------------------------------
// TLB::TLB
//      Allocate and initialize the hardware translation lookaside buffer.
//	Initially, all the entries in the table are invalid!
//----------------------------------------------------------------------

TLB::TLB()
{
    int i;

    table = new TLBEntry[TLBSize];
    for (i = 0; i < TLBSize; i++)
	table[i].valid = FALSE;
}

//----------------------------------------------------------------------
// TLB::~TLB
//      De-allocate the hardware translation lookaside buffer.
//----------------------------------------------------------------------

TLB::~TLB()
{
    delete table;
}

//----------------------------------------------------------------------
// TLB::ReadMem
//      Read "size" (1, 2, or 4) bytes of virtual memory at "addr" into 
//	the location pointed to by "value".
//
//   	Returns FALSE if the translation step from virtual to physical memory
//   	failed.
//
//	"addr" -- the virtual address to read from
//	"size" -- the number of bytes to read (1, 2, or 4)
//	"value" -- the place to write the result
//----------------------------------------------------------------------

bool
TLB::ReadMem(int addr, int size, int* value)
{
    DEBUG('a', "Reading VA 0x%x, size %d\n", addr, size);
    
    int physicalAddress;
    
    if (!Translate(addr, &physicalAddress, size, FALSE))
	return FALSE;
    switch (size) {
      case 1:
	*value = machine->mainMemory[physicalAddress];
	break;
	
      case 2:
	*value = *(unsigned short *) &machine->mainMemory[physicalAddress];
	break;
	
      case 4:
	*value = *(unsigned int *) &machine->mainMemory[physicalAddress];
	break;

      default: ASSERT(FALSE);
    }
    
    DEBUG('a', "\tvalue read = %8.8x\n", *value);
    return (TRUE);
}

//----------------------------------------------------------------------
// TLB::WriteMem
//      Write "size" (1, 2, or 4) bytes of the contents of "value" into
//	virtual memory at location "addr".
//
//   	Returns FALSE if the translation step from virtual to physical memory
//   	failed.
//
//	"addr" -- the virtual address to write to
//	"size" -- the number of bytes to be written (1, 2, or 4)
//	"value" -- the data to be written
//----------------------------------------------------------------------

bool
TLB::WriteMem(int addr, int size, int value)
{
    DEBUG('a', "Writing VA 0x%x, size %d, value 0x%x\n", addr, size, value);
    
    int physicalAddress;
    
    if (!Translate(addr, &physicalAddress, size, TRUE))
	return (FALSE);
    switch (size) {
      case 1:
	machine->mainMemory[physicalAddress] = (unsigned char) (value & 0xff);
	break;

      case 2:
	*(unsigned short *) &machine->mainMemory[physicalAddress]
		= (unsigned short) (value & 0xffff);
	break;
      
      case 4:
	*(unsigned int *) &machine->mainMemory[physicalAddress]
		= (unsigned int) value;
	break;
	
      default: ASSERT(FALSE);
    }
    
    return (TRUE);
}

//----------------------------------------------------------------------
// TLB::Translate
// 	Translate a virtual address into a physical address, using 
//	the TLB.  Check for alignment and all sorts of other errors, and 
//	set the use/dirty bits in the TLB entry.
//
// Returns FALSE if there was an error.
// Params:
//	"virtAddr" -- the virtual address to translate
//	"physAddr" -- the place to store the physical address
//	"size" -- the amount of memory being read or written
// 	"writing" -- if TRUE, check the "read-only" bit in the TLB
//----------------------------------------------------------------------
bool
TLB::Translate(int virtAddr, int* physAddr, int size, bool writing)
{
    int i;

    DEBUG('a', "\tTranslate 0x%x, %s: ", virtAddr, writing ? "write" : "read");
    
// first, calculate the virtual page number, and offset within the page,
// from the virtual address
    unsigned int vpn = (unsigned) virtAddr / PageSize;
    unsigned int offset = (unsigned) virtAddr % PageSize;
    unsigned int pageFrame;

// check for alignment errors
    if (((size == 4) && (virtAddr & 0x3)) || ((size == 2) && (offset & 0x1))) {
	DEBUG('a', "alignment problem at %d, size %d!\n", virtAddr, size);
	machine->RaiseException(AddressErrorException, virtAddr);
	return FALSE;
    }
    
    for (i = 0; i < TLBSize; i++) {
    	if (table[i].valid && (table[i].virtualPage == vpn)) {	// FOUND!
	    if (table[i].readOnly && writing) {		// trying to write
							// to a read-only page
		DEBUG('a', "%d mapped read-only at %d in TLB!\n", virtAddr, i);
		machine->RaiseException(ReadOnlyException, virtAddr);
		return FALSE;
	    }
            pageFrame = table[i].physicalPage;
// if the pageFrame is too big, there is something really wrong! 
// Most likely, an invalid translation was loaded into the TLB. 
    	    if (pageFrame >= NumPhysPages) { 
		DEBUG('a', "*** frame %d > %d!\n", pageFrame, NumPhysPages);
		machine->RaiseException(BusErrorException, virtAddr);
		return (FALSE);
	    }
    	    table[i].use = TRUE;		// set the use, dirty bits
    	    if (writing)
		table[i].dirty = TRUE;
    	    *physAddr = pageFrame * PageSize + offset;
    	    ASSERT((*physAddr >= 0) && ((*physAddr + size) <= MemorySize));
    	    DEBUG('a', "phys addr = 0x%x\n", *physAddr);
	    return TRUE;
 	}
    }
// Translation not found!

    DEBUG('a', "*** no valid TLB entry found for this virtual page!\n");
    machine->RaiseException(NoTLBEntryException, virtAddr);
    return (FALSE);
}
