// tlb.h 
//	Data structures for the software-managed translation lookaside
//	buffer (TLB).  The TLB stores a set of entries of the form:
//	<virtual page #, physical page #>.
//
//	On every memory reference, the hardware uses the memory address
//	as a key to checks the TLB to see if there is a corresponding
//	physical page #.  If so, it accesses that memory location.
//	If not, it traps to software with an exception. 
//
//	In practice, the TLB is much smaller than the amount of physical
//	memory (16 entries is common on a machine that has 1000's of
//	pages).  Thus, there must also be a backup translation scheme
//	(such as page tables), but the hardware doesn't need to know
//	anything at all about that.
//
//	Note that the contents of the TLB are specific to an address space.
//	If the address space changes, so does the contents of the TLB!
//
// DO NOT CHANGE -- part of the machine emulation
//
// Copyright (c) 1992-1993 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#ifndef TLB_H
#define TLB_H

#include "copyright.h"
#include "utility.h"

// The following class defines an entry in the translation lookaside buffer.
// Each entry defines a mapping from one virtual page to one physical page.
// In addition, there are some extra bits for access control (valid and 
// read-only) and some bits for usage information (use and dirty).

class TLBEntry {
  public:
    int virtualPage;  	// The page number in real memory.
    int physicalPage;  	// The page number in real memory.
    bool valid;         // If this bit is set, the translation is ignored.
			// (In other words, the entry hasn't been initialized.)
    bool readOnly;	// If this bit is set, the user program is not allowed
			// to modify the contents of the page.
    bool use;           // This bit is set by the hardware every time the
			// page is referenced or modified.
    bool dirty;         // This bit is set by the hardware every time the
			// page is modified.
};

// The following class defines the translation lookaside buffer.
// It consists of a table of entries, each of which maps one virtual page 
// to one physical page.
//
// NOTE: the routines here should only be called by the hardware emulation.
// The Nachos kernel can access the TLB, set or clear an entry, etc.
//
// The implementations of ReadMem/WriteMem/Translate reflect what the
// hardware does on the MIPS workstation.

class TLB {
  public:
    TLB();				// initialize the TLB
    ~TLB();				// de-allocate the TLB

    // Read or write 1, 2, or 4 bytes of virtual memory (at addr).  
    // Return FALSE if the translation couldn't be found.
    bool ReadMem(int addr, int size, int* value);
    bool WriteMem(int addr, int size, int value);
    
    // Translate an address using the TLB, and check for alignment.
    // Set the use and dirty bits in the TLB entry appropriately,
    // and return FALSE if the translation couldn't be found.
    bool Translate(int virtAddr, int* physAddr, int size, bool writing);

    TLBEntry *table;		// the table of entries in the TLB
};

#endif
