


// synch.cc -- Thread synchronization routines -- semaphores, locks 
//   and condition variables.
//
// Semaphores disable/enable interrupts for mutual exclusive 
//   access to the queue of threads waiting for the semaphore.
//
// Copyright (c) 1992 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include "copyright.h"

#include "synch.h"
#include "utility.h"
#include "thread.h"
#include "scheduler.h"
#include "interrupt.h"
#include "system.h"

// Semaphores ---------------------------------------------------------- 

// Initialize a new Semaphore.
Semaphore::Semaphore(char* semaphoreName, int val)
{
    name = semaphoreName;
    value = val;
    queue = new List;
}

// Block until semaphore value > 0; then decrement it and return.
void
Semaphore::P ()
{
    IntStatus oldLevel = interrupt->SetLevel(IntOff);
    
    // Checks to see if the semaphore is available.
    if (value == 0) { 	// semaphore not available
	queue->Append((void *)currentThread);
	currentThread->Sleep();
    } else 		 // semaphore available, consume its value
        value--;
    
    (void) interrupt->SetLevel(oldLevel);
}

// Increment semaphore value; wake up a waiting thread, if any.
void
Semaphore::V ()
{
    Thread *thread;
    IntStatus oldLevel = interrupt->SetLevel(IntOff);

    thread = (Thread *)queue->Remove();
    if (thread != NULL)	   // make thread ready, consuming the V immediately
	scheduler->ReadyToRun(thread);
    else
        value++;
    (void) interrupt->SetLevel(oldLevel);
}

// Dummy functions -- so we can compile our later assignments 
// Note -- without a correct implementation of Condition::Wait(), 
// the test case in HW5 won't work!

Lock::Lock(char* lockName) 
{
#ifdef CHANGED
  mutex = new Semaphore (lockName, 1);
  name = lockName;
#endif
}


void Lock::Acquire() 
{
#ifdef CHANGED
  mutex->P();
  holder = currentThread;
#endif
}

bool Lock::isOwner(Thread* thread)
{
#ifdef CHANGED
  if (holder == thread)
    return TRUE;
  else return FALSE;
#endif
}

void Lock::Release() 
{
#ifdef CHANGED
  ASSERT (holder == currentThread);
  mutex->V();
#endif
}


Condition::Condition(char* conditionName) 
{ 
#ifdef CHANGED
  name = conditionName;
  queue = new List;
#endif
}


void Condition::Wait(Lock* conditionLock) 
{ 
#ifdef CHANGED
  interrupt->SetLevel(IntOff);
  ASSERT(conditionLock->isOwner(currentThread));
  queue->Append((void *)currentThread);
  conditionLock->Release();
  currentThread->Sleep();
  conditionLock->Acquire();
#endif
}


void Condition::Signal(Lock* conditionLock) 
{ 
#ifdef CHANGED
  IntStatus oldLevel = interrupt->SetLevel(IntOff);
  ASSERT(conditionLock->isOwner(currentThread));
  Thread* thread = (Thread *)queue->Remove();
  if (thread != NULL)
    scheduler->ReadyToRun(thread);
  (void) interrupt->SetLevel(oldLevel);
#endif
}


void Condition::Broadcast(Lock* conditionLock) 
{ 
#ifdef CHANGED
  IntStatus oldLevel = interrupt->SetLevel(IntOff);
  ASSERT(conditionLock->isOwner(currentThread));
  Thread *thread;

  do {
    thread = (Thread *)queue->Remove();
    if (thread != NULL)
      scheduler->ReadyToRun(thread);
  } while (thread != NULL);

  (void) interrupt->SetLevel(oldLevel);
#endif
}


