/* switch.h
 *  Definitions needed for implementing context switching.
 *
 *  Context switching is inherently machine dependent, since
 *  the registers to be saved, how to set up an initial
 *  call frame, etc, are all specific to a processor architecture.
 *
 *  This file currently supports the DEC MIPS and SUN SPARC architectures.
 */

/*
 Copyright (c) 1992-1993 The Regents of the University of California.
 All rights reserved.  See copyright.h for copyright notice and limitation 
 of liability and disclaimer of warranty provisions.
 */

#ifndef SWITCH_H
#define SWITCH_H

#ifndef HOST_DOS
#include "copyright.h"
#else
#include "copyrght.hpp"
#endif

#ifdef HOST_MIPS

/* Registers that must be saved during a context switch. 
 * These are the offsets from the beginning of the Thread object, 
 * in bytes, used in switch.s
 */
#define SP 0
#define S0 4
#define S1 8
#define S2 12
#define S3 16
#define S4 20
#define S5 24
#define S6 28
#define S7 32
#define FP 36
#define PC 40

/* To fork a thread, we set up its saved register state, so that
 * when we switch to the thread, it will start running in ThreadRoot.
 *
 * The following are the initial registers we need to set up to
 * pass values into ThreadRoot (for instance, containing the procedure
 * for the thread to run).  The first set is the registers as used
 * by ThreadRoot; the second set is the locations for these initial
 * values in the Thread object -- used in Thread::AllocateStack().
 */

#define InitialPC   s0
#define InitialArg  s1
#define WhenDonePC  s2
#define StartupPC   s3

#define PCState     (PC/4-1)
#define FPState     (FP/4-1)
#define InitialPCState  (S0/4-1)
#define InitialArgState (S1/4-1)
#define WhenDonePCState (S2/4-1)
#define StartupPCState  (S3/4-1)

#endif  // HOST_MIPS

#ifdef HOST_SPARC

/* Registers that must be saved during a context switch.  See comment above. */ 
#define I0 4
#define I1 8
#define I2 12
#define I3 16
#define I4 20
#define I5 24
#define I6 28
#define I7 32

/* Aliases used for clearing code.  */
#define FP I6
#define PC I7

/* Registers for ThreadRoot.  See comment above. */
#define InitialPC       %o0
#define InitialArg      %o1
#define WhenDonePC      %o2
#define StartupPC       %o3

#define PCState         (PC/4-1)
#define InitialPCState  (I0/4-1)
#define InitialArgState (I1/4-1)
#define WhenDonePCState (I2/4-1)
#define StartupPCState  (I3/4-1)
#endif  // HOST_SPARC

#ifdef HOST_SNAKE

/* Registers that must be saved during a context switch.  See comment above. */ 
#define   SP   0
#define   S0   4
#define   S1   8
#define   S2   12
#define   S3   16
#define   S4   20
#define   S5   24
#define   S6   28
#define   S7   32
#define   S8   36
#define   S9   40
#define   S10  44
#define   S11  48
#define   S12  52
#define   S13  56
#define   S14  60
#define   S15  64
#define   PC   68

/* Registers for ThreadRoot.  See comment above. */
#define InitialPC       %r3     /* S0 */
#define InitialArg      %r4
#define WhenDonePC      %r5
#define StartupPC       %r6

#define PCState         (PC/4-1)
#define InitialPCState  (S0/4-1)
#define InitialArgState (S1/4-1)
#define WhenDonePCState (S2/4-1)
#define StartupPCState  (S3/4-1)
#endif  // HOST_SNAKE

#ifdef HOST_DOS

/* The new stack setup looks like this:
                         int Array Index 
StartUpPCState  Segment     18
StartUpPCState  Offset      17
WhenDonePCState Segment     16
WhenDonePCState Offset      15
InitialPCState  Segment     14
InitialPCState  Offset      13
InitialArgState High word   12
InitialArgState Low word    11
ThreadRoot      Segment     10
ThreadRoot      Offset      9
bp                          8
ds                          7
es                          6
ax                          5
bx                          4
cx                          3
dx                          2
di                          1
si                          0
****************************************/

/* These definitions are used in Thread::StackAllocate. */
#define StartupPCState      0   
#define InitialPCState      1
#define InitialArgState     2
#define WhenDonePCState     3

#endif // HOST_DOS


#endif // SWITCH_H
